@@ -96,7 +96,6 @@ describe "Literal Regexps" do
     /./.match("\0").to_a.should == ["\0"]
   end
 
-
   it "supports | (alternations)" do
     /a|b/.match("a").to_a.should == ["a"]
   end
@@ -115,7 +114,7 @@ describe "Literal Regexps" do
     /foo.(?<=\d)/.match("fooA foo1").to_a.should == ["foo1"]
   end
 
-  ruby_bug "#13671", ""..."3.3" do # https://bugs.ruby-lang.org/issues/13671
+  ruby_bug "#13671", ""..."3.4" do # https://bugs.ruby-lang.org/issues/13671
     it "handles a lookbehind with ss characters" do
       r =  Regexp.new("(?<!dss)", Regexp::IGNORECASE)
       r.should =~ "✨"
@@ -161,26 +160,6 @@ describe "Literal Regexps" do
     pattern.should_not =~ 'T'
   end
 
-  escapable_terminators =  ['!', '"', '#', '%', '&', "'", ',', '-', ':', ';', '@', '_', '`']
-
-  it "supports escaping characters when used as a terminator" do
-    escapable_terminators.each do |c|
-      ref = "(?-mix:#{c})"
-      pattern = eval("%r" + c + "\\" + c + c)
-      pattern.to_s.should == ref
-    end
-  end
-
-  it "treats an escaped non-escapable character normally when used as a terminator" do
-    all_terminators = [*("!".."/"), *(":".."@"), *("[".."`"), *("{".."~")]
-    special_cases = ['(', '{', '[', '<', '\\', '=', '~']
-    (all_terminators - special_cases - escapable_terminators).each do |c|
-      ref = "(?-mix:\\#{c})"
-      pattern = eval("%r" + c + "\\" + c + c)
-      pattern.to_s.should == ref
-    end
-  end
-
   it "support handling unicode 9.0 characters with POSIX bracket expressions" do
     char_lowercase = "\u{104D8}" # OSAGE SMALL LETTER A
     /[[:lower:]]/.match(char_lowercase).to_s.should == char_lowercase
